# -*- coding: utf-8 -*-

"""Top-level package for cdp-scrapers."""

__author__ = "Jackson Maxfield Brown"
__email__ = "jmaxfieldbrown@gmail.com"
# Do not edit this string manually, always use bumpversion
# Details in CONTRIBUTING.md
__version__ = "0.3.3"


def get_module_version():
    return __version__
